@@ -33,6 +33,7 @@
 #include "buffer.h"
 #include "chunk.h"
 #include "fdevent.h"
+#include "http_header.h"
 #include "log.h"
 #include "sock_addr.h"
 
@@ -314,7 +315,7 @@ static void gw_proc_check_enable(gw_host * const host, gw_proc * const proc, log
     log_error(errh, __FILE__, __LINE__,
       "gw-server re-enabled: %s %s %hu %s",
       proc->connection_name->ptr, host->host->ptr, host->port,
-      host->unixsocket->ptr ? host->unixsocket->ptr : "");
+      host->unixsocket && host->unixsocket->ptr ? host->unixsocket->ptr : "");
 }
 
 static void gw_proc_waitpid_log(const gw_host * const host, const gw_proc * const proc, log_error_st * const errh, const int status) {
@@ -1029,7 +1030,7 @@ static void gw_restart_dead_procs(gw_host * const host, log_error_st * const err
             }
 
             if (proc->state != PROC_STATE_DIED) break;
-            /* fall through *//*(we have a dead proc now)*/
+            __attribute_fallthrough__/*(we have a dead proc now)*/
 
         case PROC_STATE_DIED:
             /* local procs get restarted by us,
@@ -1890,7 +1891,7 @@ static handler_t gw_write_request(gw_handler_ctx * const hctx, request_st * cons
             hctx->reconnects = 0;
             break;
         }
-        /* fall through */
+        __attribute_fallthrough__
     case GW_STATE_CONNECT_DELAYED:
         if (hctx->state == GW_STATE_CONNECT_DELAYED) { /*(not GW_STATE_INIT)*/
             int socket_error = fdevent_connect_status(hctx->fd);
@@ -1905,7 +1906,7 @@ static handler_t gw_write_request(gw_handler_ctx * const hctx, request_st * cons
         gw_proc_connect_success(hctx->host, hctx->proc, hctx->conf.debug, r);
 
         gw_set_state(hctx, GW_STATE_PREPARE_WRITE);
-        /* fall through */
+        __attribute_fallthrough__
     case GW_STATE_PREPARE_WRITE:
         /* ok, we have the connection */
 
@@ -1929,7 +1930,7 @@ static handler_t gw_write_request(gw_handler_ctx * const hctx, request_st * cons
 
         fdevent_fdnode_event_add(hctx->ev, hctx->fdn, FDEVENT_IN|FDEVENT_RDHUP);
         gw_set_state(hctx, GW_STATE_WRITE);
-        /* fall through */
+        __attribute_fallthrough__
     case GW_STATE_WRITE:
         if (!chunkqueue_is_empty(&hctx->wb)) {
             log_error_st * const errh = r->conf.errh;
@@ -2265,7 +2266,7 @@ static handler_t gw_recv_response_error(gw_handler_ctx * const hctx, request_st
               "socket: %s for %s?%.*s, closing connection",
               (long long)hctx->wb.bytes_out, proc->connection_name->ptr,
               r->uri.path.ptr, BUFFER_INTLEN_PTR(&r->uri.query));
-        } else {
+        } else if (!light_btst(r->resp_htags, HTTP_HEADER_UPGRADE)) {
             log_error(r->conf.errh, __FILE__, __LINE__,
               "response already sent out, but backend returned error on "
               "socket: %s for %s?%.*s, terminating connection",
