@@ -5,7 +5,6 @@
 
 #include "first.h"
 
-#include "base.h"
 #include "log.h"
 
 #include <sys/types.h>
@@ -27,6 +26,8 @@
 #endif
 #endif
 
+time_t log_epoch_secs = 0;
+
 int log_clock_gettime_realtime (struct timespec *ts) {
       #ifdef HAVE_CLOCK_GETTIME
 	return clock_gettime(CLOCK_REALTIME, ts);
@@ -42,100 +43,38 @@ int log_clock_gettime_realtime (struct timespec *ts) {
 }
 
 /* retry write on EINTR or when not all data was written */
-ssize_t write_all(int fd, const void* buf, size_t count) {
-	ssize_t written = 0;
+ssize_t write_all(int fd, const void * const buf, size_t count) {
+    ssize_t written = 0;
 
-	while (count > 0) {
-		ssize_t r = write(fd, buf, count);
-		if (r < 0) {
-			switch (errno) {
-			case EINTR:
-				/* try again */
-				break;
-			default:
-				/* fail - repeating probably won't help */
-				return -1;
-			}
-		} else if (0 == r) {
-			/* really shouldn't happen... */
-			errno = EIO;
-			return -1;
-		} else {
-			force_assert(r <= (ssize_t) count);
-			written += r;
-			buf = r + (char const*) buf;
-			count -= r;
-		}
-	}
+    for (ssize_t wr; count > 0; count -= wr, written += wr) {
+        wr = write(fd, (const char *)buf + written, count);
+        if (wr > 0) continue;
 
-	return written;
-}
+        if (wr < 0 && errno == EINTR) { wr = 0; continue; } /* try again */
+        if (0 == wr) errno = EIO; /* really shouldn't happen... */
+        return -1; /* fail - repeating probably won't help */
+    }
 
-/* lowercase: append space, uppercase: don't */
-static void log_buffer_append_printf(buffer *out, const char *fmt, va_list ap) {
-	for(; *fmt; fmt++) {
-		int d;
-		char *s;
-		buffer *b;
-		off_t o;
-
-		switch(*fmt) {
-		case 'S':           /* string */
-		case 's':           /* string */
-			s = va_arg(ap, char *);
-			buffer_append_string_c_escaped(out, s, (NULL != s) ? strlen(s) : 0);
-			break;
-		case 'B':           /* buffer */
-		case 'b':           /* buffer */
-			b = va_arg(ap, buffer *);
-			buffer_append_string_c_escaped(out, CONST_BUF_LEN(b));
-			break;
-		case 'D':           /* int */
-		case 'd':           /* int */
-			d = va_arg(ap, int);
-			buffer_append_int(out, d);
-			break;
-		case 'O':           /* off_t */
-		case 'o':           /* off_t */
-			o = va_arg(ap, off_t);
-			buffer_append_int(out, o);
-			break;
-		case 'X':           /* int (hex) */
-		case 'x':           /* int (hex) */
-			d = va_arg(ap, int);
-			buffer_append_string_len(out, CONST_STR_LEN("0x"));
-			buffer_append_uint_hex(out, d);
-			break;
-		case '(':
-		case ')':
-		case '<':
-		case '>':
-		case ',':
-		case ' ':
-			buffer_append_string_len(out, fmt, 1);
-			break;
-		}
-
-		if (*fmt >= 'a') { /* 's' 'b' 'd' 'o' 'x' */
-			buffer_append_string_len(out, CONST_STR_LEN(" "));
-		}
-	}
+    return written;
 }
 
 static int log_buffer_prepare(const log_error_st *errh, const char *filename, unsigned int line, buffer *b) {
+	static time_t tlast;
+	static char tstr[20]; /* 20-chars needed for "%Y-%m-%d %H:%M:%S" */
+	static size_t tlen;
 	switch(errh->errorlog_mode) {
 	case ERRORLOG_PIPE:
 	case ERRORLOG_FILE:
 	case ERRORLOG_FD:
 		if (-1 == errh->errorlog_fd) return -1;
 		/* cache the generated timestamp */
-		if (*errh->last_ts != *errh->cur_ts) {
-			*errh->last_ts = *errh->cur_ts;
-			buffer_clear(errh->tb);
-			buffer_append_strftime(errh->tb, "%Y-%m-%d %H:%M:%S", localtime(errh->cur_ts));
+		if (tlast != log_epoch_secs) {
+			tlast = log_epoch_secs;
+			tlen = strftime(tstr, sizeof(tstr),
+			                "%Y-%m-%d %H:%M:%S", localtime(&tlast));
 		}
 
-		buffer_copy_buffer(b, errh->tb);
+		buffer_copy_string_len(b, tstr, tlen);
 		buffer_append_string_len(b, CONST_STR_LEN(": ("));
 		break;
 	case ERRORLOG_SYSLOG:
@@ -166,61 +105,17 @@ static void log_write(const log_error_st *errh, buffer *b) {
 	}
 }
 
-int log_error_write(server *srv, const char *filename, unsigned int line, const char *fmt, ...) {
-	const log_error_st *errh = srv->errh;
-	buffer *b = errh->b;
-	if (-1 == log_buffer_prepare(errh, filename, line, b)) return 0;
 
-	va_list ap;
-	va_start(ap, fmt);
-	log_buffer_append_printf(b, fmt, ap);
-	va_end(ap);
-
-	log_write(errh, b);
-
-	return 0;
-}
-
-int log_error_write_multiline_buffer(server *srv, const char *filename, unsigned int line, buffer *multiline, const char *fmt, ...) {
-	const log_error_st *errh = srv->errh;
-	buffer *b = errh->b;
-	va_list ap;
-	size_t prefix_len;
-	char *pos, *end, *current_line;
-
-	if (buffer_string_is_empty(multiline)) return 0;
-
-	if (-1 == log_buffer_prepare(errh, filename, line, b)) return 0;
-
-	va_start(ap, fmt);
-	log_buffer_append_printf(b, fmt, ap);
-	va_end(ap);
-
-	prefix_len = buffer_string_length(b);
-
-	current_line = pos = multiline->ptr;
-	end = multiline->ptr + buffer_string_length(multiline);
-
-	for ( ; pos <= end ; ++pos) {
-		switch (*pos) {
-		case '\n':
-		case '\r':
-		case '\0': /* handles end of string */
-			if (current_line < pos) {
-				/* truncate to prefix */
-				buffer_string_set_length(b, prefix_len);
-
-				buffer_append_string_len(b, current_line, pos - current_line);
-				log_write(errh, b);
-			}
-			current_line = pos + 1;
-			break;
-		default:
-			break;
-		}
-	}
-
-	return 0;
+static void
+log_buffer_append_encoded (buffer * const b,
+                           const char * const s, const size_t n)
+{
+    size_t i;
+    for (i = 0; i < n && ' ' <= s[i] && s[i] <= '~'; ++i) ;/*(ASCII isprint())*/
+    if (i == n)
+        buffer_append_string_len(b, s, n);  /* common case; nothing to encode */
+    else
+        buffer_append_string_c_escaped(b, s, n);
 }
 
 
@@ -263,14 +158,14 @@ log_buffer_vprintf (buffer * const b,
 
 
 static void
-log_error_va_list_impl (const log_error_st * const errh,
+log_error_va_list_impl (log_error_st * const errh,
                         const char * const filename,
                         const unsigned int line,
                         const char * const fmt, va_list ap,
                         const int perr)
 {
     const int errnum = errno;
-    buffer * const b = errh->b;
+    buffer * const b = &errh->b;
     if (-1 == log_buffer_prepare(errh, filename, line, b)) return;
     log_buffer_vprintf(b, fmt, ap);
     if (perr) {
@@ -283,7 +178,7 @@ log_error_va_list_impl (const log_error_st * const errh,
 
 
 void
-log_error(const log_error_st * const errh,
+log_error(log_error_st * const errh,
           const char * const filename, const unsigned int line,
           const char *fmt, ...)
 {
@@ -295,7 +190,7 @@ log_error(const log_error_st * const errh,
 
 
 void
-log_perror (const log_error_st * const errh,
+log_perror (log_error_st * const errh,
             const char * const filename, const unsigned int line,
             const char * const fmt, ...)
 {
@@ -306,17 +201,47 @@ log_perror (const log_error_st * const errh,
 }
 
 
+void
+log_error_multiline_buffer (log_error_st * const restrict errh,
+                            const char * const restrict filename,
+                            const unsigned int line,
+                            const buffer * const restrict multiline,
+                            const char * const restrict fmt, ...)
+{
+    if (multiline->used < 2) return;
+
+    const int errnum = errno;
+    buffer * const b = &errh->b;
+    if (-1 == log_buffer_prepare(errh, filename, line, b)) return;
+
+    va_list ap;
+    va_start(ap, fmt);
+    log_buffer_vprintf(b, fmt, ap);
+    va_end(ap);
+
+    const size_t prefix_len = buffer_string_length(b);
+    const char * const end = multiline->ptr + multiline->used - 2;
+    const char *pos = multiline->ptr-1, *current_line;
+    do {
+        pos = strchr(current_line = pos+1, '\n');
+        if (!pos)
+            pos = end;
+        buffer_string_set_length(b, prefix_len); /* truncate to prefix */
+        log_buffer_append_encoded(b, current_line, pos - current_line);
+        log_write(errh, b);
+    } while (pos < end);
+
+    errno = errnum;
+}
+
+
 log_error_st *
-log_error_st_init (time_t *cur_ts_ptr, time_t *last_ts_ptr)
+log_error_st_init (void)
 {
     log_error_st *errh = calloc(1, sizeof(log_error_st));
     force_assert(errh);
     errh->errorlog_fd = STDERR_FILENO;
     errh->errorlog_mode = ERRORLOG_FD;
-    errh->b = buffer_init();
-    errh->tb = buffer_init();
-    errh->cur_ts = cur_ts_ptr;
-    errh->last_ts = last_ts_ptr;
     return errh;
 }
 
@@ -325,7 +250,6 @@ void
 log_error_st_free (log_error_st *errh)
 {
     if (NULL == errh) return;
-    buffer_free(errh->tb);
-    buffer_free(errh->b);
+    free(errh->b.ptr);
     free(errh);
 }
