@@ -356,7 +356,7 @@ static int server_oneshot_read_cq(connection *con, chunkqueue *cq, off_t max_byt
     /* temporary set con->fd to oneshot_fd (fd input) rather than outshot_fdout
      * (lighttpd generally assumes operation on sockets, so this is a kludge) */
     int fd = con->fd;
-    con->fd = oneshot_fd;
+    con->fd = oneshot_fdn->fd;
     int rc = oneshot_read_cq(con, cq, max_bytes);
     con->fd = fd;
 
@@ -915,6 +915,17 @@ static int server_graceful_state_bg (server *srv) {
     return 1;
 }
 
+__attribute_cold__
+__attribute_noinline__
+static void server_graceful_shutdown_maint (server *srv) {
+    if (oneshot_fd) {
+        /* permit keep-alive on one-shot connections until graceful_expire_ts */
+        if (!srv->graceful_expire_ts) return;
+        if (srv->graceful_expire_ts >= log_epoch_secs) return;
+    }
+    connection_graceful_shutdown_maint(srv);
+}
+
 __attribute_cold__
 static void server_graceful_state (server *srv) {
 
@@ -927,11 +938,13 @@ static void server_graceful_state (server *srv) {
             if (srv->graceful_expire_ts)
                 srv->graceful_expire_ts += log_epoch_secs;
         }
-        connection_graceful_shutdown_maint(srv);
+        server_graceful_shutdown_maint(srv);
     }
 
-    if (!oneshot_fd
-        && (2 == srv->sockets_disabled || 3 == srv->sockets_disabled)) return;
+    if (2 == srv->sockets_disabled || 3 == srv->sockets_disabled) {
+        if (oneshot_fd) graceful_restart = 0;
+        return;
+    }
 
     log_error(srv->errh,__FILE__,__LINE__,"[note] graceful shutdown started");
 
@@ -1125,7 +1138,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 			return -1;
 		}
 		graceful_shutdown = 1;
-		srv->sockets_disabled = 1;
+		srv->sockets_disabled = 2;
 		srv->srvconf.dont_daemonize = 1;
 		if (srv->srvconf.pid_file) buffer_clear(srv->srvconf.pid_file);
 		if (srv->srvconf.max_worker) {
@@ -1496,7 +1509,8 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 			log_error(srv->errh, __FILE__, __LINE__, "Opening errorlog failed. Going down.");
 			return -1;
 		}
-		log_error(srv->errh, __FILE__, __LINE__, "server started (" PACKAGE_DESC ")");
+		if (!oneshot_fd)
+			log_error(srv->errh, __FILE__, __LINE__, "server started (" PACKAGE_DESC ")");
 	}
 
 	if (HANDLER_GO_ON != plugins_call_set_defaults(srv)) {
@@ -1571,7 +1585,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 					switch (fdevent_reaped_logger_pipe(pid)) {
 					  default: break;
 					  case -1: if (!timer) alarm((timer = 5));
-						   /* fall through */
+						   __attribute_fallthrough__
 					  case  1: continue;
 					}
 					/** 
@@ -1821,7 +1835,8 @@ static void server_handle_sigalrm (server * const srv, time_t min_ts, time_t las
 				/* reset global/aggregate rate limit counters */
 				config_reset_config_bytes_sec(srv->config_data_base);
 				/* if graceful_shutdown, accelerate cleanup of recently completed request/responses */
-				if (graceful_shutdown && !srv_shutdown) connection_graceful_shutdown_maint(srv);
+				if (graceful_shutdown && !srv_shutdown)
+					server_graceful_shutdown_maint(srv);
 				connection_periodic_maint(srv, min_ts);
 }
 
@@ -1955,7 +1970,7 @@ int main (int argc, char **argv) {
             if (2 == graceful_shutdown) { /* value 2 indicates idle timeout */
                 log_error(srv->errh, __FILE__, __LINE__,
                   "server stopped after idle timeout");
-            } else {
+            } else if (!oneshot_fd) {
               #ifdef HAVE_SIGACTION
                 log_error(srv->errh, __FILE__, __LINE__,
                   "server stopped by UID = %d PID = %d",
